/**
 *
 * QRcode class library 0.50beta10<BR>
 * (c)2003-2005 Y.Swetake<BR>
 * This version supports QRcode model2 version 1-40.<BR>
 * Some functions are not supported.<BR>
 *<BR>
 * @author Y.Swetake
 * @version 0.50beta10
 *
 * ported from Java to C++ 2005 by NAKATA H. <webmaster@nakata-jp.org>
 *
 *
 */
#include <cstdio>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <ios>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define	VERSION_TO_SIZE(ver)	(4 * (ver) + 17)
#define	XY2INDEX(x,y)		(((x) & 0xff) + ((y) & 0xff) * modules1Side)
#define	BITS_IN_BYTE		8
#define	TRUE			(-1)
#define	FALSE			0

class Qrcode{
 private:
  char qrcodeErrorCorrect;
  char qrcodeEncodeMode;
  char *qrcodeDataPath;
  int qrcodeVersion;

  int qrcodeStructureappendN;
  int qrcodeStructureappendM;
  int qrcodeStructureappendParity;

 public:
  int	modules1Side;
  Qrcode() {
    qrcodeErrorCorrect = 'M';
    qrcodeEncodeMode = 'B';
    qrcodeVersion = 0;
		
    qrcodeStructureappendN = 0;
    qrcodeStructureappendM = 0;
    qrcodeStructureappendParity = 0;

    pattern = NULL;
    codewords = NULL;
    res = NULL;
    qrcodeDataPath = strdup("qrcode_data");
  }

  ~Qrcode() {
    if ( pattern != NULL ) {
      delete pattern;
    }
    if ( codewords != NULL ) {
      delete codewords;
    }
    if ( res != NULL ) {
      delete res;
    }
    free(qrcodeDataPath);
  }

  int setQrcodeDataPath(char *path) {
    struct stat buf;
    stat( path, &buf );
    if ( !S_ISDIR(buf.st_mode) ) {
      return (FALSE);
    }
	 
    free( qrcodeDataPath );
    qrcodeDataPath = strdup(path);
      
    return (TRUE);
  }

  /**
   *顼٥ꤷޤ
   *@param ecc-顼٥('L','M','Q','H')
   *
   */
  void setQrcodeErrorCorrect(char ecc) {
    qrcodeErrorCorrect = ecc;
  }

  /**
   *ꤵƤ륨顼٥ޤ
   *@return 顼٥('L','M','Q','H')
   *
   */
  char getQrcodeErrorCorrect() {
    return qrcodeErrorCorrect;
  }

  /**
   *ꤵƤСޤ
   *@return С 0400ξϼưꡣ
   *
   */
  int getQrcodeVersion() {
    return qrcodeVersion;
  }

  /**
   *Сꤷޤ
   *0ꤹȼưˤʤޤ
   *@param version  040
   *
   */
  void setQrcodeVersion(int ver) {
    if (ver>=0 && ver<=40){
      qrcodeVersion = ver;
    }
  }

  /**
   *󥳡ɥ⡼ɤꤷޤ
   *'N':⡼  'A':ѿ⡼ɡ¾:8bit byte⡼
   *@param encMode 󥳡ɥ⡼('N','A' or other)
   *
   */
  void setQrcodeEncodeMode(char encMode) {
    qrcodeEncodeMode = encMode;
  }

  /**
   *ꤵƤ륨󥳡ɥ⡼ɤޤ
   *@return 󥳡ɥ⡼ ('N','A' or other)
   *
   */
  char getQrcodeEncodeMode() {
    return qrcodeEncodeMode;
  }
  
  /**
   *ϢϢΥ᥽åɤǤ
   *(ƳǤ)
   *
   */
  void setStructureappend(int m,int n,int p) {
    if (1<n && n<=16 && 0<m && m<=16 && 0<=p && p<=255){
      qrcodeStructureappendM = m;
      qrcodeStructureappendN = n;
      qrcodeStructureappendParity = p;
    }
  }

  /**
   *ϢѤѥƥ򻻽Фޤ
   *(ƳǤ)
   */
  int calStructureappendParity(char *originaldata){
    int originaldataLength;
    int structureappendParity;
    
    originaldataLength = strlen(originaldata);
    
    if (originaldataLength <= 1){
      structureappendParity = -1;
    } else {
      structureappendParity = 0;
      for ( int i=0; i<originaldataLength; i++ ) {
	structureappendParity ^= (originaldata[i] & 0xFF);
      }
    }
    return structureappendParity;
  }

  /**
   *Ϳ줿ǡ󤫤QRɥ󥳡ɥǡ
   *boolean󼡸֤ޤ
   *@param data 󥳡ɤǡ
   *@returnQRɥ󥳡ɥǡ
   */
  char	*pattern;
  char	*calQrcode(const char *qrcodeData, const int dataLength) {
    static int codewordNumPlusAlnum[] = {0,0,0,0,0,0,0,0,0,0,
					 2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
					 4,4,4,4,4,4,4,4,4,4,4,4,4,4};
    static int codewordNumPlusNumeric[] = {0,0,0,0,0,0,0,0,0,0,
					   2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
					   4,4,4,4,4,4,4,4,4,4,4,4,4,4};
    static int codewordNumPlus8bit[] = {0,0,0,0,0,0,0,0,0,0,
					8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
					8,8,8,8,8,8,8,8,8,8,8,8,8,8};
    static int codewordNumPlusKanji[] = {0,0,0,0,0,0,0,0,0,0,
					2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
					4,4,4,4,4,4,4,4,4,4,4,4,4,4};
    int dataCounter;

    int dataValue[dataLength+32];
    char dataBits[dataLength+32];

    if (dataLength <= 0) {
      return NULL;
    }

    if (qrcodeStructureappendN <= 1) {
      dataCounter = 0;
    } else {
      dataValue[0] = 3;
      dataBits[0] = 4;

      dataValue[1] = qrcodeStructureappendM - 1;
      dataBits[1] = 4;

      dataValue[2] = qrcodeStructureappendN - 1;
      dataBits[2] = 4;

      dataValue[3] = qrcodeStructureappendParity;
      dataBits[3] = 8;

      dataCounter = 4;
    }
    dataBits[dataCounter] = 4;

    /*  --- determine encode mode --- */

    int *codewordNumPlus;
    int codewordNumCounterValue;

    switch (qrcodeEncodeMode) {
      /* ---- alphanumeric mode ---  */
    case 'A':
      codewordNumPlus = codewordNumPlusAlnum;
      dataValue[dataCounter++] = 2;
      dataValue[dataCounter] = dataLength;
      dataBits[dataCounter] = 9;
      codewordNumCounterValue = dataCounter;

      dataCounter++;
      for (int i=0; i<dataLength; i++) {
	char chr = (char)qrcodeData[i];
	char chrValue = 0;
	if ( '0' <= chr && chr < 58) {
	  chrValue = chr - '0';
	} else if ( 65 <= chr && chr < 91) {
	  chrValue = chr - 55;
	} else {
	  if (chr == 32) {chrValue = 36;}
	  if (chr == 36) {chrValue = 37;}
	  if (chr == 37) {chrValue = 38;}
	  if (chr == 42) {chrValue = 39;}
	  if (chr == 43) {chrValue = 40;}
	  if (chr == 45) {chrValue = 41;}
	  if (chr == 46) {chrValue = 42;}
	  if (chr == 47) {chrValue = 43;}
	  if (chr == 58) {chrValue = 44;}
	}
	if ((i % 2) == 0) {
	  dataValue[dataCounter] = chrValue;
	  dataBits[dataCounter] = 6;
	} else {
	  dataValue[dataCounter] = dataValue[dataCounter] * 45 + chrValue;
	  dataBits[dataCounter] = 11;
	  if ( i < dataLength - 1 ) {
	    dataCounter++;
	  }
	}
      }
      dataCounter++;
      break;

      /* kanji */
    case 'K':
      codewordNumPlus = codewordNumPlusKanji;
      dataValue[dataCounter++] = 8;
      dataValue[dataCounter] = dataLength / 2;
      dataBits[dataCounter] = 8;
      codewordNumCounterValue = dataCounter;

      dataCounter++;
      for (int i = 0; i < dataLength; i += 2, dataCounter++) {
	int	c;
	c = (unsigned char)qrcodeData[i];
	c -= ( c < 0xe0 ) ? 0x81 : 0xc1;
	dataValue[dataCounter] = c * 0xc0;
	c = (unsigned char)qrcodeData[i + 1] - 0x40;
	dataValue[dataCounter] += c;
	dataBits[dataCounter] = 13;
      }
      break;

      /* ---- numeric mode ---- */
    case 'N':
      codewordNumPlus = codewordNumPlusNumeric;
      dataValue[dataCounter++] = 1;
      dataValue[dataCounter] = dataLength;
      
      dataBits[dataCounter] = 10;   /* #version 1-9*/ 
      codewordNumCounterValue = dataCounter;

      for (int i=0; i < dataLength; i++) {
	if ((i % 3) == 0) {
	  dataCounter++;
	  dataValue[dataCounter] = 0;
	  dataBits[dataCounter] = 1;
	}
	dataValue[dataCounter] = dataValue[dataCounter] * 10
	  + (qrcodeData[i] - '0');
	dataBits[dataCounter] += 3;
      }
      dataCounter++;
      break;

      /* ---- 8bit byte ---- */
    default:
      codewordNumPlus = codewordNumPlus8bit;
      dataValue[dataCounter++] = 4;
      dataValue[dataCounter] = dataLength;
      dataBits[dataCounter] = 8;   /* #version 1-9 */
      codewordNumCounterValue = dataCounter;

      dataCounter++;
      
      for (int i=0; i < dataLength; i++, dataCounter++) {
	dataValue[dataCounter] = (qrcodeData[i] & 0xFF);
	dataBits[dataCounter] = BITS_IN_BYTE;
      }
      
      break;
    }

    int totalDataBits = 0;
    for (int i=0; i < dataCounter; i++) {
      totalDataBits += dataBits[i];
    }

    int ec;
    switch (qrcodeErrorCorrect){
    case 'L':
      ec = 1;
      break;
    case 'Q':
      ec = 3;
      break;
    case 'H':
      ec = 2;
      break;
    default:
      ec = 0;
    }


    int maxDataBitsArray[][41] = {
      {0,128,224,352,512,688,864,992,1232,1456,1728,
       2032,2320,2672,2920,3320,3624,4056,4504,5016,5352,
       5712,6256,6880,7312,8000,8496,9024,9544,10136,10984,
       11640,12328,13048,13800,14496,15312,15936,16816,17728,18672},

      {0,152,272,440,640,864,1088,1248,1552,1856,2192,
       2592,2960,3424,3688,4184,4712,5176,5768,6360,6888,
       7456,8048,8752,9392,10208,10960,11744,12248,13048,13880,
       14744,15640,16568,17528,18448,19472,20528,21616,22496,23648},

      {0,72,128,208,288,368,480,528,688,800,976,
       1120,1264,1440,1576,1784,2024,2264,2504,2728,3080,
       3248,3536,3712,4112,4304,4768,5024,5288,5608,5960,
       6344,6760,7208,7688,7888,8432,8768,9136,9776,10208},

      {0,104,176,272,384,496,608,704,880,1056,1232,
       1440,1648,1952,2088,2360,2600,2936,3176,3560,3880,
       4096,4544,4912,5312,5744,6032,6464,6968,7288,7880,
       8264,8920,9368,9848,10288,10832,11408,12016,12656,13328}
    };

    int maxDataBits = 0;

    if (qrcodeVersion > 0) {
      maxDataBits = maxDataBitsArray[ec][qrcodeVersion];
    } else {
      /* auto version select */
      for (qrcodeVersion = 1; qrcodeVersion <= 40; qrcodeVersion++) {
	maxDataBits = maxDataBitsArray[ec][qrcodeVersion];
	if (maxDataBits >= totalDataBits + codewordNumPlus[qrcodeVersion]) {
	  break;
	}
      }
      if ( qrcodeVersion >= 41 ) {
	std::cerr << "data size too large\n";
	return NULL;
      }
    }

    totalDataBits += codewordNumPlus[qrcodeVersion];
    dataBits[codewordNumCounterValue] += codewordNumPlus[qrcodeVersion];

    int maxCodewordsArray[] = {0,26,44,70,100,134,172,196,242,292,
			       346,404,466,532,581,655,733,815,901,991,
			       1085,1156,1258,1364,1474,1588,1706,1828,1921,2051,
			       2185,2323,2465,2611,2761,2876,3034,3196,3362,3532,
			       3706};

    int maxCodewords = maxCodewordsArray[qrcodeVersion];

    int matrixRemainBit[] = {0,0,7,7,7,7,7,0,0,0,0,0,0,0,3,3,3,3,3,3,3,
			     4,4,4,4,4,4,4,3,3,3,3,3,3,3,0,0,0,0,0,0};

    /* ---- read version ECC data file */

    int byte_num = matrixRemainBit[qrcodeVersion]
      + (maxCodewords * BITS_IN_BYTE);

    char matrixX[byte_num];
    char matrixY[byte_num];
    char maskArray[byte_num];
    char formatInformationX2[15];
    char formatInformationY2[15];
    char rsEccCodewords;
    char rsBlockOrder[128];
    memset(rsBlockOrder, 0, 128);

    char	filename[256];
    try {
      sprintf( filename, "%s/qrv%d_%d.dat",
	       qrcodeDataPath, qrcodeVersion, ec );

      std::ifstream bis;
      bis.open(filename, std::ios_base::in | std::ios_base::binary);

      bis.read(matrixX, byte_num);
      bis.read(matrixY, byte_num);
      bis.read(maskArray, byte_num);
      bis.read(formatInformationX2, 15);
      bis.read(formatInformationY2, 15);
      bis.read(&rsEccCodewords, 1);
      bis.read(rsBlockOrder, 128);
      bis.close();
    }
    catch(...) {
      std::cerr << "file " << filename << " access error.\n";
      return NULL;
    }
    
    char rsBlockOrderLength = 1;
    for (int i = 1; i < 128; i++) {
      if (rsBlockOrder[i] == 0) {
	rsBlockOrderLength = i;
	break;
      }
    }

    char formatInformationX1[] = {0,1,2,3,4,5,7,8,8,8,8,8,8,8,8};
    char formatInformationY1[] = {8,8,8,8,8,8,8,8,7,5,4,3,2,1,0};

    int maxDataCodewords = maxDataBits / BITS_IN_BYTE;

    /* -- read frame data  -- */

    modules1Side = VERSION_TO_SIZE(qrcodeVersion);
    int matrixTotalBits = modules1Side * modules1Side;
    char frameData[matrixTotalBits + modules1Side];

    try {
      sprintf(filename, "%s/qrvfr%d.dat", qrcodeDataPath, qrcodeVersion);

      std::ifstream	bis;
      bis.open(filename, std::ios_base::in | std::ios_base::binary);
      bis.read(frameData, matrixTotalBits + modules1Side);
      bis.close();
    }
    catch (...) {
      std::cerr << "file " << filename << " access error.\n";
      return NULL;
    }

    /*  --- set terminator */

    if (totalDataBits <= maxDataBits - 4) {
      dataValue[dataCounter] = 0;
      dataBits[dataCounter] = 4;
    } else if (totalDataBits < maxDataBits) {
      dataValue[dataCounter] = 0;
      dataBits[dataCounter] = maxDataBits - totalDataBits;
    } else if (totalDataBits > maxDataBits) {
      std::cerr << "overflow\n";
      return NULL;
    }
    divideDataBy8Bits(dataValue, dataBits, dataCounter + 1, maxDataCodewords);
    char *codewords = calculateRSECC(rsEccCodewords,
				     rsBlockOrder, rsBlockOrderLength,
				     maxDataCodewords, maxCodewords);

    /* ---- flash matrix */

    char matrixContent[modules1Side * modules1Side];

    for (int i=0; i < modules1Side; i++) {
      for (int j=0; j < modules1Side; j++) {
        matrixContent[XY2INDEX(j,i)] = 0;
      }
    }

    /* --- attach data */
    for (int i=0; i < maxCodewords; i++) {
      char codeword_i = codewords[i];
      for (int j=7; j >= 0; j--){
        int codewordBitsNumber = (i*8) + j;

        matrixContent[ XY2INDEX(matrixX[codewordBitsNumber],
				matrixY[codewordBitsNumber])]
	  = ( (255 * (codeword_i & 1)) ^ maskArray[codewordBitsNumber] ); 

        codeword_i = (codeword_i & 0xFF) >> 1;
      }
    }

    for (int matrixRemain = matrixRemainBit[qrcodeVersion];
	 matrixRemain > 0; matrixRemain--) {
      int remainBitTemp = matrixRemain + (maxCodewords * 8) - 1;
      matrixContent[ XY2INDEX(matrixX[remainBitTemp], matrixY[remainBitTemp])]
	= 255 ^ maskArray[remainBitTemp];
    }

    /* --- mask select --- */
    char maskNumber = selectMask(matrixContent, modules1Side,
				 matrixRemainBit[qrcodeVersion]
				 + maxCodewords * 8);
    char maskContent = 1 << maskNumber;

    /* --- format information --- */

    char formatInformationValue = ec << 3 | maskNumber;

    char *formatInformationArray[] = {"101010000010010", "101000100100101",
				      "101111001111100", "101101101001011",
				      "100010111111001", "100000011001110",
				      "100111110010111", "100101010100000",
				      "111011111000100", "111001011110011",
				      "111110110101010", "111100010011101",
				      "110011000101111", "110001100011000",
				      "110110001000001", "110100101110110",
				      "001011010001001", "001001110111110",
				      "001110011100111", "001100111010000",
				      "000011101100010", "000001001010101",
				      "000110100001100", "000100000111011",
				      "011010101011111", "011000001101000",
				      "011111100110001", "011101000000110",
				      "010010010110100", "010000110000011",
				      "010111011011010", "010101111101101"};

    for (int i=0; i < 15; i++) {
      char content = formatInformationArray[(int)formatInformationValue][i];

      matrixContent[XY2INDEX(formatInformationX1[i], formatInformationY1[i])]
	= (content == '0') ? 0 : 255;
      matrixContent[XY2INDEX(formatInformationX2[i], formatInformationY2[i])]
	= (content == '0') ? 0 : 255;
    }

    if ( pattern != NULL ) {
      delete pattern;
    }
    pattern = new char[modules1Side * modules1Side];

    int c = 0;
    for (int i=0; i < modules1Side; i++, c++) {
      for (int j=0; j < modules1Side; j++, c++) {
        if ((matrixContent[XY2INDEX(j,i)] & maskContent) != 0
	    || frameData[c] == 49) {
	  pattern[XY2INDEX(j,i)] = 0;
        } else {
	  pattern[XY2INDEX(j,i)] = 255;
        }
      }
    }

    return pattern;
}

private:
  char	*codewords;
  int	codewordsLength;
  void	divideDataBy8Bits(const int data[] ,const char bits[],
			  const int length, const int maxDataCodewords) {
    /* divide Data By 8bit and add padding char */
    int l2;
    int remainingBits;
    int max = 0;
    int buffer;
    int bufferBits;
    int flag;

    for (int i=0; i < length; i++) {
      max += bits[i];
    }
    l2 = (max - 1) / BITS_IN_BYTE + 1;

    if ( codewords != NULL ) {
      delete codewords;
    }
    codewords = new char[maxDataCodewords];

    for (int i=0; i < l2; i++) {
      codewords[i] = 0;
    }
    codewordsLength = 0;
    remainingBits = BITS_IN_BYTE;
    for (int i=0; i < length && bits[i] > 0; i++) {
      buffer = data[i];
      bufferBits = bits[i];

      while (bufferBits > 0) {
	if (remainingBits > bufferBits) {
	  codewords[codewordsLength] =
	    (codewords[codewordsLength] << bufferBits) | buffer;
	  remainingBits -= bufferBits;
	  break;		// now bufferBits is 0
	}
	bufferBits -= remainingBits;
	codewords[codewordsLength]
	  = (codewords[codewordsLength] << remainingBits)
	  | (buffer >> bufferBits);
	buffer &= 0xffffffff >> (32 - bufferBits);
	codewordsLength++;
	remainingBits = BITS_IN_BYTE;
      }
    }
    if (remainingBits != BITS_IN_BYTE) {
      codewords[codewordsLength++] <<= remainingBits;
    }

    for ( flag = TRUE; codewordsLength < maxDataCodewords; flag = !flag ) {
      codewords[codewordsLength++] = flag ? -20 : 17;
    }
  }

  char	*res;
  char *calculateRSECC(const char rsEccCodewords, 
		       const char rsBlockOrder[], const int rsBlockOrderLength,
		       const int maxDataCodewords, const int maxCodewords) {
    
    char	rsCalTableArray[256][rsEccCodewords];
    char	filename[256];
    try {
      sprintf(filename, "%s/rsc%d.dat", qrcodeDataPath, rsEccCodewords);

      std::ifstream	bis;
      bis.open(filename, std::ios_base::in | std::ios_base::binary);

      for (int i = 0; i < 256; i++) {
	bis.read(&(rsCalTableArray[i][0]), rsEccCodewords);
      }
      bis.close();
    }
    catch (...) {
      std::cerr << "file " << filename << " access error.\n";
      return NULL;
    }

    /* ---- RS-ECC prepare */

    int	i, j;

    char	*rsTemp[rsBlockOrderLength];
    res = NULL;

    if ( res != NULL ) {
      delete res;
    }
    res = new char[maxCodewords];
    memcpy(res, codewords, codewordsLength);

    for( i = 0; i < rsBlockOrderLength; i++) {
      rsTemp[i] = new char[(rsBlockOrder[i] & 0xFF) - rsEccCodewords];
    }

    int rsBlockNumber = 0;
    for ( j = 0, i = 0; i < maxDataCodewords; i++, j++ ) {
      if ( j >= (rsBlockOrder[rsBlockNumber]& 0xFF) - rsEccCodewords ) {
	j = 0;
	rsBlockNumber++;
      }
      rsTemp[rsBlockNumber][j] = codewords[i];
    }

    /* ---  RS-ECC main --- */

    for ( rsBlockNumber = 0;
	  rsBlockNumber < rsBlockOrderLength; rsBlockNumber++ ) {
      int	rsTempDataLength = (rsBlockOrder[rsBlockNumber] & 0xff)
	- rsEccCodewords;
      int	tmp = (rsTempDataLength > rsEccCodewords)
	? rsTempDataLength : rsEccCodewords;
      char	rsTempData[tmp];
      memcpy(rsTempData ,rsTemp[rsBlockNumber], rsTempDataLength );

      int rsCodewords = (rsBlockOrder[rsBlockNumber] & 0xFF);
      int rsDataCodewords = rsCodewords - rsEccCodewords;

      for ( j = rsDataCodewords; j > 0; j-- ) {
	char first = rsTempData[0];
	if ( first != 0 ) {
	  for (int k = 0; k < rsTempDataLength - 1; k++ ) {
	    rsTempData[k] = rsTempData[k+1];
	  }
	  rsTempDataLength--;
	  if ( rsTempDataLength > rsEccCodewords ) {
	    for (int k = 0; k < rsEccCodewords; k++ ) {
	      rsTempData[k] ^= rsCalTableArray[(first & 0xff)][k];
	    }
	  } else {
	    int	k;
	    for ( k = 0; k < rsTempDataLength; k++ ) {
	      rsTempData[k] ^= rsCalTableArray[(first & 0xff)][k];
	    }
	    for ( ; k < rsEccCodewords; k++ ) {
	      rsTempData[k] = rsCalTableArray[(first & 0xff)][k];
	    }
	    rsTempDataLength = rsEccCodewords;
	  }
	} else if ( rsEccCodewords < rsTempDataLength ) {
	  for (int k = 0; k < rsTempDataLength - 1; k++ ) {
	    rsTempData[k] = rsTempData[k+1];
	  }
	  rsTempDataLength--;
	} else {
	  int	k;
	  for (k = 0; k < rsTempDataLength - 1; k++ ) {
	    rsTempData[k] = rsTempData[k+1];
	  }
	  for ( ; k < rsEccCodewords; k++ ) {
	    rsTempData[k] = 0;
	  }
	  rsTempDataLength = rsEccCodewords;
	}
      }
      memcpy(&(res[codewordsLength + rsBlockNumber * rsEccCodewords]),
	      rsTempData, rsEccCodewords);
    }
    for( i = 0; i < rsBlockOrderLength; i++) {
      delete rsTemp[i];
    }
    return res;
  }

  int selectMask(const char matrixContent[], const int modules1Side,
		 const int maxCodewordsBitWithRemain) {
    int d1[] = {0,0,0,0,0,0,0,0};
    int d2[] = {0,0,0,0,0,0,0,0};
    int d3[] = {0,0,0,0,0,0,0,0};
    int d4[] = {0,0,0,0,0,0,0,0};

    int d2And = 0;
    int d2Or = 0;
    int d4Counter[] = {0,0,0,0,0,0,0,0};

    for (int y = 0; y < modules1Side; y++ ) {
      int xData[] = {0,0,0,0,0,0,0,0};
      int yData[] = {0,0,0,0,0,0,0,0};
      int xD1Flag[] = {FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};
      int yD1Flag[] = {FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};

      for (int x = 0; x < modules1Side; x++) {
	if ( x > 0 && y > 0 ) {
	  d2And = matrixContent[XY2INDEX(x,y)] & matrixContent[XY2INDEX(x-1,y)]
	    & matrixContent[XY2INDEX(x,y-1)] & matrixContent[XY2INDEX(x-1,y-1)]
	    & 0xFF;

	  d2Or = (matrixContent[XY2INDEX(x,y)] | matrixContent[XY2INDEX(x-1,y)]
		  | matrixContent[XY2INDEX(x,y-1)]
		  | matrixContent[XY2INDEX(x-1,y-1)]) & 0xFF;
	}
	
	for(int maskNumber = 0; maskNumber < 8; maskNumber++ ) {
	  xData[maskNumber] = ((xData[maskNumber] & 63) << 1)
	    | ((matrixContent[XY2INDEX(x,y)] >> maskNumber) & 1);
 
	  yData[maskNumber] = ((yData[maskNumber] & 63) << 1)
	    | ((matrixContent[XY2INDEX(y,x)] >> maskNumber) & 1);

	  if ( (matrixContent[XY2INDEX(x,y)] & (1 << maskNumber)) != 0 ) {
	    d4Counter[maskNumber]++;
	  }

	  if ( xData[maskNumber] == 93 ) {
	    d3[maskNumber] += 40;
	  }
	  if ( yData[maskNumber] == 93 ) {
	    d3[maskNumber] += 40;
	  }

	  if ( x > 0 && y > 0 ) {
	    if ( ((d2And & 1) != 0) || ((d2Or & 1) == 0) ) {
	      d2[maskNumber] += 3;
	    }
	    d2And >>= 1;
	    d2Or >>= 1;
	  }

	  if ( ((xData[maskNumber] & 0x1F) == 0)
	       || ((xData[maskNumber] & 0x1F) == 0x1F) ) {
	    if ( x > 3 ) {
	      if ( xD1Flag[maskNumber] ) {
		d1[maskNumber]++;
	      } else {
		d1[maskNumber] += 3;
		xD1Flag[maskNumber] = TRUE;
	      }
	    }
	  } else {
	    xD1Flag[maskNumber] = FALSE;
	  }
	  if ( ((yData[maskNumber] & 0x1F) == 0)
	       || ((yData[maskNumber] & 0x1F) == 0x1F) ) {
	    if ( x > 3 ) {
	      if ( yD1Flag[maskNumber] ) {
		d1[maskNumber]++;
	      } else {
		d1[maskNumber] += 3;
		yD1Flag[maskNumber] = TRUE;
	      }
	    }
	  } else {
	    yD1Flag[maskNumber] = FALSE;
	  }
	}
      }
    }
    int minValue = 0;
    int res = 0;
    int d4Value[] = {90,80,70,60,50,40,30,20,10,0,
		     0,10,20,30,40,50,60,70,80,90,90};
    for (int maskNumber = 0; maskNumber < 8; maskNumber++) {
      d4[maskNumber] = d4Value[(int)((20 * d4Counter[maskNumber])
				     / maxCodewordsBitWithRemain)];

      int demerit = d1[maskNumber] + d2[maskNumber]
	+ d3[maskNumber] + d4[maskNumber];
      
      if ( demerit < minValue || maskNumber == 0) {
	res = maskNumber;
	minValue = demerit;
      }
    }
    return (res);
  }

/*--- class end ---*/
};
